import numpy as np
from calngUtils import misc as utils, shmem_utils
from karabo.bound import (
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
    ChannelMetaData,
    Dims,
    Encoding,
    Epochstamp,
    Hash,
    ImageData,
    Schema,
    State,
    Timestamp,
    Trainstamp,
)
from TrainMatcher import TrainMatcher

from . import scenes, schemas
from ._version import version as deviceVersion


@KARABO_CLASSINFO("Gotthard2Assembler", deviceVersion)
class Gotthard2Assembler(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            OVERWRITE_ELEMENT(expected)
            .key("availableScenes")
            .setNewDefaultValue(["overview", "trainMatcherScene"])
            .commit(),

            OUTPUT_CHANNEL(expected)
            .key("output")
            .dataSchema(schemas.preview_schema(wrap_image_in_imagedata=False))
            .commit(),

            STRING_ELEMENT(expected)
            .key("geometryDevice")
            .displayedName("[Unused] Geometry device")
            .description(
                "Assembly of 25 μm GOTTHARD-II does not require geometry per se. Just "
                "make sure to, in the sources list, list the master module first and "
                "the slave module second."
            )
            .assignmentOptional()
            .defaultValue("")
            .commit(),

            STRING_ELEMENT(expected)
            .key("assemblyMode")
            .displayedName("Assembly mode")
            .description(
                "Previews for 25 μm GOTTHARD-II are generally generated by "
                "interleaving the preview outputs of the two constituent 50 μm "
                "modules. However, the frame sum previews is temporal (sum across all "
                "pixels per frame), so this preview should be the sum of the two "
                "constituent previews. Additionally, previews are either 1D (regular "
                "ones) or 2D (streak previews) and the latter need special image data "
                "wrapping. This 'assembler' can therefore either interleave (1D or 2D) "
                "or sum - 'auto' means it will guess which to do based on the primary "
                "input source name."
            )
            .options("auto,interleave1d,interleave2d,sum")
            .assignmentOptional()
            .defaultValue("auto")
            .commit(),
        )

    def initialization(self):
        super().initialization()
        sources = self.get("sources")
        if len(sources) != 2:
            self.log.WARN("Must have exactly two GOTTHARD-II sources for 25 μm mode")
            self.updateState(State.ERROR)
        self._primary_source, self._secondary_source = [
            row["source"].partition("@")[0] for row in self.get("sources")
        ]

        # figure out assembly mode and set handler
        if self.get("assemblyMode") == "auto":
            _, _, source_output = self._primary_source.partition(":")
            if source_output.lower().endswith("streak"):
                self.set("assemblyMode", "interleave2d")
            elif source_output.lower().endswith("sums"):
                self.set("assemblyMode", "sum")
            else:
                self.set("assemblyMode", "interleave1d")
        mode = self.get("assemblyMode")
        if mode == "interleave1d":
            self._do_the_assembly = self._interleave_1d
        elif mode == "interleave2d":
            self._do_the_assembly = self._interleave_2d
            # we may need to re-inject output channel to satisfy GUI :D
            schema_update = Schema()
            (
                OUTPUT_CHANNEL(schema_update)
                .key("output")
                .dataSchema(
                    schemas.preview_schema(wrap_image_in_imagedata=True)
                )
                .commit(),
            )
            self.updateSchema(schema_update)
            self.output = self._ss.getOutputChannel("output")
        else:
            self._do_the_assembly = self._sum_1d

        self._shmem_handler = shmem_utils.ShmemCircularBufferReceiver()
        self._interleaving_buffer = np.ma.empty(0, dtype=np.float32)
        self._wrap_in_imagedata = False
        self.start()

    def requestScene(self, params):
        scene_name = params.get("name", default="")
        if scene_name == "trainMatcherScene":
            params["name"] = "scene"
            return super().requestScene(params)
        if scene_name != "overview":
            self.log.INFO(
                f"Unexpected scene name {scene_name} requested, "
                "will just return overview"
            )
        payload = Hash("name", scene_name, "success", True)
        payload["data"] = scenes.gotthard2_assembler_overview(
            device_id=self.getInstanceId(),
            schema=self.getFullSchema(),
        )
        self.reply(
            Hash(
                "type",
                "deviceScene",
                "origin",
                self.getInstanceId(),
                "payload",
                payload,
            )
        )

    def on_matched_data(self, train_id, sources):
        if (
            missing_sources := {self._primary_source, self._secondary_source}
            - sources.keys()
        ):
            self.log.WARN(
                f"Missing preview source(s): {missing_sources}, skipping train"
            )
            return
        for (data, _) in sources.values():
            self._shmem_handler.dereference_shmem_handles(data)

        # data hash format varies between dataOutput and preview.output*
        if sources[self._primary_source][0].has("data.adc"):
            data_1 = sources[self._primary_source][0]["data.adc"]
            data_2 = sources[self._secondary_source][0]["data.adc"]
        else:
            data_1 = sources[self._primary_source][0]["image.data"]
            data_2 = sources[self._secondary_source][0]["image.data"]
        if sources[self._primary_source][0].has("image.mask"):
            mask_1 = sources[self._primary_source][0]["image.mask"]
            mask_2 = sources[self._secondary_source][0]["image.mask"]
        else:
            mask_1 = False
            mask_2 = False

        meta = ChannelMetaData(
            f"{self.getInstanceId()}:output",
            Timestamp(Epochstamp(), Trainstamp(train_id)),
        )
        output_hash = self._do_the_assembly(data_1, mask_1, data_2, mask_2)
        self.output.write(output_hash, meta, copyAllData=False)
        self.output.update(safeNDArray=True)

        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

    def _interleave_1d(self, data_1, mask_1, data_2, mask_2):
        image_1 = np.ma.masked_array(data=data_1, mask=mask_1)
        image_2 = np.ma.masked_array(data=data_2, mask=mask_2)

        # now to figure out the interleaving
        out_shape = utils.interleaving_buffer_shape(data_1.shape, 2, 0)
        if self._interleaving_buffer.shape != out_shape:
            self._interleaving_buffer = np.ma.masked_array(
                np.empty(shape=out_shape, dtype=np.float32),
                mask=False,
            )
        utils.set_on_axis(self._interleaving_buffer, image_1, np.index_exp[0::2], 0)
        utils.set_on_axis(self._interleaving_buffer, image_2, np.index_exp[1::2], 0)
        # TODO: replace this part with preview friend
        self._interleaving_buffer.mask |= ~np.isfinite(self._interleaving_buffer.data)

        return Hash(
            "image.data",
            self._interleaving_buffer.data,
            "image.mask",
            self._interleaving_buffer.mask,
        )

    def _interleave_2d(self, data_1, mask_1, data_2, mask_2):
        data_1 = data_1.getData()
        data_2 = data_2.getData()
        mask_1 = mask_1.getData()
        mask_2 = mask_2.getData()

        image_1 = np.ma.masked_array(data=data_1, mask=mask_1)
        image_2 = np.ma.masked_array(data=data_2, mask=mask_2)

        out_shape = utils.interleaving_buffer_shape(data_1.shape, 2, 1)
        if self._interleaving_buffer.shape != out_shape:
            self._interleaving_buffer = np.ma.masked_array(
                np.empty(shape=out_shape, dtype=np.float32),
                mask=False,
            )
        utils.set_on_axis(self._interleaving_buffer, image_1, np.index_exp[0::2], 1)
        utils.set_on_axis(self._interleaving_buffer, image_2, np.index_exp[1::2], 1)
        # TODO: replace this part with preview friend
        self._interleaving_buffer.mask |= ~np.isfinite(self._interleaving_buffer.data)

        return Hash(
            "image.data",
            ImageData(
                self._interleaving_buffer.data,
                Dims(*self._interleaving_buffer.shape),
                Encoding.GRAY,
                bitsPerPixel=32,
            ),
            "image.mask",
            ImageData(
                self._interleaving_buffer.mask,
                Dims(*self._interleaving_buffer.shape),
                Encoding.GRAY,
                bitsPerPixel=32,
            ),
        )

    def _sum_1d(self, data_1, mask_1, data_2, mask_2):
        image_1 = np.ma.masked_array(data=data_1, mask=mask_1)
        image_2 = np.ma.masked_array(data=data_2, mask=mask_2)
        # don't bother with self._interleaving_buffer
        res = image_1 + image_2

        return Hash(
            "image.data",
            res.data,
            "image.mask",
            res.mask,
        )
