import threading

import numpy as np
from calngUtils import trackers
from karabo.middlelayer import (
    AccessLevel,
    AccessMode,
    Bool,
    Configurable,
    DaqPolicy,
    Device,
    Double,
    EncodingType,
    Hash,
    Image,
    ImageData,
    InputChannel,
    Node,
    OutputChannel,
    Overwrite,
    Slot,
    State,
    String,
    UInt32,
    VectorDouble,
    VectorInt32,
    VectorString,
    get_property,
    slot,
)

from . import scenes, utils


def image_data_node():
    return Image(
        ImageData(np.zeros((0, 0), dtype=np.float32), encoding=EncodingType.GRAY),
    )


class PreviewSettingsNode(Configurable):
    downsamplingMethod = String(
        options=["nanmax", "nanmean", "nanmin", "nanmedian"], defaultValue="nanmax"
    )
    downsamplingFactor = UInt32(options=[1, 2, 4, 8], defaultValue=1)
    flipX = Bool(defaultValue=True, accessMode=AccessMode.RECONFIGURABLE)
    flipY = Bool(defaultValue=True, accessMode=AccessMode.RECONFIGURABLE)


class HistogramSettingsNode(Configurable):
    @Slot(displayedName="Reset bins")
    async def resetBins(self):
        parent = self.get_root()
        with parent._bin_lock:
            parent._bins = None
            parent._means = None

    resetBinsOnRoiChange = Bool(
        defaultValue=True,
        accessMode=AccessMode.RECONFIGURABLE,
        displayedName="Reset on ROI change",
    )
    automaticallyExpandRange = Bool(
        defaultValue=True,
        accessMode=AccessMode.RECONFIGURABLE,
        displayedName="Auto-expand",
    )

    @Double(defaultValue=0, accessMode=AccessMode.RECONFIGURABLE, displayedName="Min")
    async def rangeMin(self, value):
        self.rangeMin = value
        await self.resetBins()

    @Double(
        defaultValue=True, accessMode=AccessMode.RECONFIGURABLE, displayedName="Max"
    )
    async def rangeMax(self, value):
        self.rangeMax = value
        await self.resetBins()

    @UInt32(
        defaultValue=10,
        maxInc=100_000,
        accessMode=AccessMode.RECONFIGURABLE,
        displayedName="#Bins",
    )
    async def numBins(self, value):
        self.numBins = value
        await self.resetBins()

    @UInt32(
        defaultValue=10,
        maxInc=100_000,
        accessMode=AccessMode.RECONFIGURABLE,
        displayedName="Window size",
    )
    async def rollingWindowSize(self, value):
        self.rollingWindowSize = value
        await self.resetBins()


class HistogramDisplayableNode(Configurable):
    x = VectorDouble()
    y = VectorDouble()
    yMean = VectorDouble()


class OutputNode(Configurable):
    # note: would be cool to not re-send image, but do it to ensure ImageData
    fullImage = image_data_node()
    zoomImage = image_data_node()
    histogram = Node(HistogramDisplayableNode)


class RoiTool(Device):
    state = Overwrite(defaultValue=State.INIT)
    output = OutputChannel(OutputNode)
    imageDataPath = String(
        defaultValue="image.data", accessMode=AccessMode.RECONFIGURABLE
    )
    maskPath = String(defaultValue="image.mask", accessMode=AccessMode.RECONFIGURABLE)
    histogram = Node(HistogramSettingsNode)
    preview = Node(PreviewSettingsNode)
    rate = Double(defaultValue=0, accessMode=AccessMode.READONLY)
    numPixelsIncluded = UInt32(
        defaultValue=0, accessMode=AccessMode.READONLY, displayedName="#Pixels counted"
    )

    def __init__(self, config):
        super().__init__(config)
        self._bins = None
        self._means = None
        self._bin_lock = threading.RLock()
        self._rate_tracker = trackers.WindowRateTracker()

    async def onInitialization(self):
        self.state = State.ON

    @VectorInt32(defaultValue=[0, 100, 0, 100])
    async def roi(self, value):
        self.roi = value
        if self.histogram.resetBinsOnRoiChange:
            await self.histogram.resetBins()

    @InputChannel()
    async def imageInput(self, data, meta):
        # TODO: handle streams without explicit mask?
        image = np.ma.masked_array(
            data=get_property(data, self.imageDataPath.value).astype(
                np.float32, copy=False
            ),
            mask=get_property(data, self.maskPath.value).astype(np.uint8, copy=False),
        )

        # TODO: make handling of extra dimension(s) configurable
        if image.ndim == 3:
            image = image[0]

        if self.preview.flipX.value:
            image = image[:, ::-1]
        if self.preview.flipY.value:
            image = image[::-1]

        x_min, x_max, y_min, y_max = self.roi.value
        x_min = max(x_min, 0)
        y_min = max(y_min, 0)
        x_max = min(x_max, image.shape[0])
        y_max = min(y_max, image.shape[1])
        if (downsampling_factor := self.preview.downsamplingFactor.value) > 1:
            x_min *= downsampling_factor
            y_min *= downsampling_factor
            x_max *= downsampling_factor
            y_max *= downsampling_factor

        # data for analysis may contain NaNs, we should filter them out
        zoomed = image[y_min:y_max, x_min:x_max]
        # TODO: still handle NaN in image?
        finite_pixel_count = zoomed.count()
        update_histogram = finite_pixel_count > 0

        if update_histogram:
            with self._bin_lock:
                if self.histogram.automaticallyExpandRange.value:
                    # device itself updating value doesn't triggers change function
                    ranges_changed = False
                    if (min_val := np.min(zoomed)) < self.histogram.rangeMin.value:
                        self.histogram.rangeMin = min_val
                        ranges_changed = True
                    if (max_val := np.max(zoomed)) > self.histogram.rangeMax.value:
                        self.histogram.rangeMax = max_val
                        ranges_changed = True
                    if ranges_changed:
                        await self.histogram.resetBins()
                if self._bins is None:
                    counts, bin_edges = np.histogram(
                        zoomed.compressed(),
                        bins=self.histogram.numBins.value,
                        range=(
                            self.histogram.rangeMin.value,
                            self.histogram.rangeMax.value,
                        ),
                    )
                    self._bins = bin_edges.astype(np.float32)
                    self._window_counts = np.zeros(
                        (
                            self.histogram.rollingWindowSize.value,
                            self.histogram.numBins.value,
                        ),
                        dtype=np.float32,
                    )
                    self._window_count_index = 0
                else:
                    counts, bin_edges = np.histogram(
                        zoomed.compressed(), bins=self._bins
                    )
                counts = counts.astype(np.float32)
                self._window_counts[self._window_count_index] = counts
                self._window_count_index = (
                    self._window_count_index + 1
                ) % self._window_counts.shape[0]

            self._rate_tracker.update()

        if update_histogram:
            (
                self.output.schema.histogram.x,
                self.output.schema.histogram.y,
            ) = _histogram_plot_helper(counts / np.sum(counts), bin_edges)
            _, self.output.schema.histogram.yMean = _histogram_plot_helper(
                np.sum(self._window_counts, axis=0) / np.sum(self._window_counts),
                bin_edges,
            )

        self.output.schema.fullImage = ImageData(
            utils.downsample_2d(
                image.filled(fill_value=0),
                downsampling_factor,
                getattr(np, self.preview.downsamplingMethod.value),
            ),
            encoding=EncodingType.GRAY,
            bitsPerPixel=32,
        )
        self.output.schema.zoomImage = ImageData(
            zoomed.filled(fill_value=0),
            encoding=EncodingType.GRAY,
            bitsPerPixel=32,
        )
        await self.output.writeData()
        self.rate = self._rate_tracker.get()
        if finite_pixel_count != self.numPixelsIncluded:
            self.numPixelsIncluded = finite_pixel_count

    availableScenes = VectorString(
        displayedName="Available scenes",
        displayType="Scenes",
        requiredAccessLevel=AccessLevel.OBSERVER,
        accessMode=AccessMode.READONLY,
        defaultValue=[
            "overview",
        ],
        daqPolicy=DaqPolicy.OMIT,
    )

    @slot
    def requestScene(self, params):
        return Hash(
            "type",
            "deviceScene",
            "origin",
            self.deviceId,
            "payload",
            Hash(
                "success",
                True,
                "name",
                "overview",
                "data",
                scenes.roitool_overview(
                    self.deviceId,
                    self.getDeviceSchema(),
                ),
            ),
        )


def _histogram_plot_helper(counts, bin_edges):
    x = np.zeros(bin_edges.size * 2)
    y = np.zeros_like(x)

    x[::2] = bin_edges
    x[1::2] = bin_edges

    y[1:-1:2] = counts
    y[2:-1:2] = counts

    return x, y
