import concurrent.futures
from timeit import default_timer

from calngUtils import shmem_utils, trackers
from karabo.bound import (
    BOOL_ELEMENT,
    DOUBLE_ELEMENT,
    KARABO_CLASSINFO,
    OVERWRITE_ELEMENT,
    UINT32_ELEMENT,
    ChannelMetaData,
    Hash,
    MetricPrefix,
    State,
    Unit,
)
from TrainMatcher import TrainMatcher

from .stacking_utils import StackingFriend
from .frameselection_utils import FrameselectionFriend
from ._version import version as deviceVersion


@KARABO_CLASSINFO("ShmemTrainMatcher", deviceVersion)
class ShmemTrainMatcher(TrainMatcher.TrainMatcher):
    @staticmethod
    def expectedParameters(expected):
        (
            # order is important for stacking, disable sorting
            OVERWRITE_ELEMENT(expected)
            .key("sortSources")
            .setNowReadOnly()
            .setNewDefaultValue(False)
            .commit(),

            BOOL_ELEMENT(expected)
            .key("enableKaraboOutput")
            .displayedName("Enable Karabo channel")
            .allowedStates(State.PASSIVE)
            .assignmentOptional()
            .defaultValue(True)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(expected)
            .key("processingTime")
            .displayedName("Processing time")
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .readOnly()
            .initialValue(0)
            .commit(),

            UINT32_ELEMENT(expected)
            .key("processingThreads")
            .allowedStates(State.PASSIVE)
            .assignmentOptional()
            .defaultValue(16)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected)
            .key("useInfiniband")
            .description(
                "If enabled, device will during initialization try to bind its data "
                "output channel (output) to its node's infiniband interface. Default "
                "interface is used if no 'ib0' interface is found."
            )
            .assignmentOptional()
            .defaultValue(True)
            .commit(),
        )
        FrameselectionFriend.add_schema(expected)
        StackingFriend.add_schema(expected)

    def __init__(self, config):
        if config.get("useInfiniband", default=True):
            from PipeToZeroMQ.utils import find_infiniband_ip

            config["output.hostname"] = find_infiniband_ip()
        self._processing_time_tracker = trackers.ExponentialMovingAverage(alpha=0.3)
        super().__init__(config)
        self.info.merge(Hash("processingTime", 0))

    def initialization(self):
        super().initialization()
        self._shmem_handler = shmem_utils.ShmemCircularBufferReceiver()
        self._stacking_friend = StackingFriend(
            self, self.get("sources"), self.get("merge")
        )
        self._frameselection_friend = FrameselectionFriend(self.get("frameSelector"))
        self._thread_pool = concurrent.futures.ThreadPoolExecutor(
            max_workers=self.get("processingThreads")
        )

        if not self.get("enableKaraboOutput"):
            # it is set already by super by default, so only need to turn off
            self.output = None

        self.start()  # Auto-start this type of matcher.

    def preReconfigure(self, conf):
        super().preReconfigure(conf)
        if conf.has("merge") or conf.has("sources"):
            self._stacking_friend.reconfigure(conf.get("sources"), conf.get("merge"))
        if conf.has("frameSelector"):
            self._frameselection_friend.reconfigure(conf.get("frameSelector"))
        if conf.has("processingThreads"):
            self._thread_pool.shutdown()
            self._thread_pool = concurrent.futures.ThreadPoolExecutor(
                max_workers=self.get("processingThreads")
            )
        if conf.has("enableKaraboOutput"):
            if conf["enableKaraboOutput"]:
                self.output = self._ss.getOutputChannel("output")
            else:
                self.output = None

    def on_matched_data(self, train_id, sources):
        ts_start = default_timer()
        frame_selection_mask = self._frameselection_friend.get_mask(sources)
        concurrent.futures.wait(
            [
                self._thread_pool.submit(
                    self._handle_source,
                    source,
                    data,
                    timestamp,
                    frame_selection_mask,
                )
                for source, (data, timestamp) in sources.items()
            ]
        )
        self._stacking_friend.process(sources, self._thread_pool)

        # karabo output
        if self.output is not None:
            for source, (data, timestamp) in sources.items():
                self.output.write(
                    data, ChannelMetaData(source, timestamp), copyAllData=False
                )
            self.output.update(safeNDArray=True)

        # karabo bridge output
        for source, (data, timestamp) in sources.items():
            self.zmq_output.write(source, data, timestamp)
        self.zmq_output.update()

        self._processing_time_tracker.update(default_timer() - ts_start)
        self.info["processingTime"] = self._processing_time_tracker.get() * 1000
        self.info["sent"] += 1
        self.info["trainId"] = train_id
        self.rate_out.update()

    def _handle_source(
        self,
        source,
        data_hash,
        timestamp,
        frame_selection_mask,
    ):
        self._shmem_handler.dereference_shmem_handles(data_hash)
        self._frameselection_friend.apply_mask(source, data_hash, frame_selection_mask)
