import copy
import functools
import json
import multiprocessing
import pathlib
import threading

import calibration_client
import h5py
import numpy as np
from calibration_client.modules import (
    Calibration,
    CalibrationConstant,
    CalibrationConstantVersion,
    Detector,
    DetectorType,
    PhysicalDetectorUnit,
)
from calngUtils import misc as utils, timing
from karabo.bound import (
    DOUBLE_ELEMENT,
    NODE_ELEMENT,
    SLOT_ELEMENT,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    VECTOR_UINT32_ELEMENT,
    Unit,
)


class ConditionNotFound(Exception):
    pass


class DetectorNotFound(Exception):
    pass


class ModuleNotFound(Exception):
    pass


class CalibrationNotFound(Exception):
    pass


class CalibrationClientConfigError(Exception):
    pass


def add_status_schema_from_enum(schema, enum_class):
    for constant in enum_class:
        constant_node = f"foundConstants.{constant.name}"
        (
            NODE_ELEMENT(schema).key(constant_node).commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.state")
            .setSpecialDisplayType("State")
            .readOnly()
            .initialValue("OFF")
            .commit(),

            SLOT_ELEMENT(schema)
            .key(f"{constant_node}.loadMostRecent")
            .displayedName("Load most recent")
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.beginValidityAt")
            .displayedName("Constant timestamp")
            .description(
                "This field displays the field 'begin_validity_at' from CalCat of the "
                "most recently found + loaded constant. This field is tied to when the "
                "constant was created and is useful to notice if some constants are "
                "out of date."
            )
            .readOnly()
            .initialValue("")
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.calibrationId")
            .displayedName("[CalCat] Calibration ID")
            .description(
                "In CalCat, each constant type is associated with a calibration ID "
                "(for example, 'Offset' has ID 1). If calibration ID is not found, "
                "something is quite wrong: missing in CalCat or issue with correction "
                "device."
            )
            .readOnly()
            .initialValue("")
            .commit(),

            VECTOR_UINT32_ELEMENT(schema)
            .key(f"{constant_node}.conditionIds")
            .displayedName("[CalCat] Condition IDs found")
            .description(
                "In CalCat, constants are found based on Physical Detector Unit and "
                "the operating condition of the detector. A condition ID represents a "
                "match on these. As operating conditions can overlap, a query may "
                "involve multiple valid condition IDs; the retrieval strategy later "
                "('closest in time') will determine which candidate constant is loaded."
            )
            .readOnly()
            .initialValue([])
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.usedConditionId")
            .displayedName("[CalCat] Condition ID used")
            .description(
                "Of the condition IDs found, this one applies to the constant that was "
                "retrieved in the end."
            )
            .readOnly()
            .initialValue("")
            .commit(),

            VECTOR_UINT32_ELEMENT(schema)
            .key(f"{constant_node}.constantIds")
            .displayedName("[CalCat] Constant IDs found")
            .description(
                "In CalCat, a 'constant' represents matching 'calibration ID', "
                "'detector type ID', and 'condition ID'. Similar to condition IDs, a "
                "query may yield multiple candidate constant IDs."
            )
            .readOnly()
            .initialValue([])
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.usedConstantId")
            .displayedName("[CalCat] Constant ID used")
            .description(
                "Of the candidate constant IDs, this one applies to the constant which "
                "was retrieved in the end."
            )
            .readOnly()
            .initialValue("")
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.constantVersionId")
            .displayedName("[CalCat] Constant version ID")
            .description(
                "Finally, a constant version represents a concrete constant that can "
                "be loaded (see descriptions of preceding fields for details on other "
                "steps). "
                "This field is editable - if for any reason a specific constant "
                "version is desired, the constant version ID (as used in CalCat) can "
                "be set here and the slot below can be called to load this particular "
                "version. Only relevant in very unusual situations."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            SLOT_ELEMENT(schema)
            .key(f"{constant_node}.overrideConstantFromVersion")
            .displayedName("Override constant version")
            .description("See description of constant version id.")
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.dataFilePath")
            .displayedName("Data file path")
            .description(
                "Where is the actual file for the constant located?"
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{constant_node}.dataSetName")
            .displayedName("Data set name")
            .description(
                "Within the actual data file, where is the constant data located?"
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"{constant_node}.constantLoadingTime")
            .displayedName("FS load time")
            .unit(Unit.SECOND)
            .description("How long it took loading this constant from the file system")
            .readOnly()
            .initialValue(0)
            .commit(),

            SLOT_ELEMENT(schema)
            .key(f"{constant_node}.overrideConstantFromFile")
            .displayedName("Override from file")
            .description("Constant from manually specified .h5 file and data set name")
            .commit(),
        )


class OperatingConditions(dict):
    # TODO: support deviation?
    def encode(self):
        return {
            "parameters": [
                {
                    "parameter_name": key,
                    "lower_deviation_value": 0.0,
                    "upper_deviation_value": 0.0,
                    "flg_available": False,
                    "value": value,
                }
                for (key, value) in self.items()
            ]
        }

    def __hash__(self):
        # this takes me back to pre-screening interview time...
        return hash(tuple(sorted(self.items())))


class BaseCalcatFriend:
    """Base class for CalCat friends - handles interacting with CalCat for the device

    A CalCat friend uses the device schema to build up parameters for CalCat queries.
    It focuses on two nodes (added by static method add_schema): constantParameters and
    foundConstants. The former is primarily used to get parameters which are (via
    condition methods - see for example dark_condition of DsscCalcatFriend) used
    to look for constants. The latter is primarily used to give user information
    about what was found.
    """

    _constant_enum_class = None  # subclass should set
    _constants_need_conditions = None  # subclass should set

    @staticmethod
    def add_schema(schema, detector_type):
        """Add elements needed by this object to device's schema (expectedSchema)

        All elements added to schema go under prefixes which should end with name of
        node which does not exist yet. To change default values and add more fields,
        extend this method in subclass.

        The constantParameters node will hold all the parameters needed to build
        constant condition dicts for querying CalCat. These values are set either
        directly on the device or via manager and this class gets them from the device
        using helper function _get_param. See for example
        AgipdCalcatFriend.dark_condition.

        The foundConstants node is used to report information about what was found in
        CalCat. This class will update the values on the device using the helper
        function _set_status. This should not need to happen in subclass methods.
        """

        (
            NODE_ELEMENT(schema)
            .key("constantParameters")
            .displayedName("Constant retrieval parameters")
            .commit(),

            NODE_ELEMENT(schema)
            .key("foundConstants")
            .displayedName("Constants retrieved")
            .commit(),
        )

        # Parameters which any detector would probably have (extend this in subclass)
        # TODO: probably switch to floating point for everything, including mem cells
        (
            STRING_ELEMENT(schema)
            .key("constantParameters.deviceMappingSnapshotAt")
            .tags("managed")
            .displayedName("Snapshot timestamp (for device mapping)")
            .description(
                "CalCat supports querying with a specific snapshot of the database. "
                "When playing back a run from the file system, this feature is useful "
                "to look up the device mapping at the time of the run. If this field "
                "is left empty, the latest device mapping is used. Date format should "
                "be 'YYYY-MM-DD' with optional time of day starting with 'T' followed "
                "by 'hh:mm:ss.mil+02:00'."
            )
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.constantVersionEventAt")
            .tags("managed")
            .displayedName("Event at timestamp (for constant version)")
            .description("TODO")
            .assignmentOptional()
            .defaultValue("")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.detectorType")
            .displayedName("Detector type name")
            .description(
                "Name of detector type in CalCat; typically has suffix '-Type'"
            )
            .readOnly()
            .initialValue(detector_type)
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.detectorTypeId")
            .readOnly()
            .initialValue("")
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.detectorName")
            .assignmentOptional()
            .defaultValue("")
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.detectorId")
            .readOnly()
            .initialValue("")
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.karaboDa")
            .assignmentOptional()
            .defaultValue("")
            .commit(),

            STRING_ELEMENT(schema)
            .key("constantParameters.moduleId")
            .readOnly()
            .initialValue("")
            .commit(),

            UINT32_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .tags("managed")
            .displayedName("Memory cells")
            .description(
                "Number of memory cells / frames per train. Relevant for burst mode."
            )
            .assignmentOptional()
            .defaultValue(1)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key("constantParameters.pixelsX")
            .displayedName("Pixels X")
            .assignmentOptional()
            .defaultValue(512)
            .commit(),

            UINT32_ELEMENT(schema)
            .key("constantParameters.pixelsY")
            .displayedName("Pixels Y")
            .assignmentOptional()
            .defaultValue(128)
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .tags("managed")
            .displayedName("Bias voltage")
            .description("Sensor bias voltage")
            .assignmentOptional()
            .defaultValue(300)
            .reconfigurable()
            .commit(),
        )

    def __init__(
        self,
        device,
        secrets_fn: pathlib.Path,
    ):
        self.device = device
        self.cached_constants = {}
        self.cached_constants_lock = threading.RLock()
        # api lock used to force queries to be sequential (SSL issue on ONC)
        self.api_lock = threading.RLock()

        if not secrets_fn.is_file():
            self.device.log_status_warn(
                f"Missing CalCat secrets file (expected {secrets_fn})"
            )
        with secrets_fn.open("r") as fd:
            calcat_secrets = json.load(fd)

        self.caldb_store = pathlib.Path(calcat_secrets["caldb_store_path"])
        if not self.caldb_store.is_dir():
            raise ValueError(f"caldb_store location '{self.caldb_store}' is not dir")

        self.device.log.INFO(f"Connecting to CalCat at {calcat_secrets['base_url']}")
        base_url = calcat_secrets["base_url"]
        self.client = calibration_client.CalibrationClient(
            client_id=calcat_secrets["client_id"],
            client_secret=calcat_secrets["client_secret"],
            user_email=calcat_secrets["user_email"],
            base_api_url=f"{base_url}/api/",
            token_url=f"{base_url}/oauth/token",
            refresh_url=f"{base_url}/oauth/token",
            auth_url=f"{base_url}/oauth/authorize",
            scope="public",
            session_token=None,
        )
        self.device.log_status_info("CalCat connection established")

    def _get_param(self, key):
        """Helper to get value from attached device schema"""
        return self.device.get(f"constantParameters.{key}")

    def _set_param(self, key, value):
        self.device.set(f"constantParameters.{key}", value)

    def _get_status(self, constant, key):
        return self.device.get(f"foundConstants.{constant.name}.{key}")

    def _set_status(self, constant, key, value):
        """Helper to update information about found constants on device"""
        self.device.set(f"foundConstants.{constant.name}.{key}", value)

    @functools.cached_property
    def detector_id(self):
        detector_name = self._get_param("detectorName")
        with self.api_lock:
            resp = Detector.get_by_identifier(self.client, detector_name)
        self._check_resp(resp, DetectorNotFound, f"Detector {detector_name} not found")
        res = resp["data"]["id"]
        self._set_param("detectorId", str(res))
        return res

    @functools.cached_property
    def detector_type_id(self):
        detector_type = self._get_param("detectorType")
        with self.api_lock:
            resp = DetectorType.get_by_name(self.client, detector_type)
        self._check_resp(
            resp, DetectorNotFound, f"Detector type {detector_type} not found"
        )
        res = resp["data"]["id"]
        self._set_param("detectorTypeId", str(res))
        return res

    @functools.cached_property
    def pdus(self):
        with self.api_lock:
            resp = PhysicalDetectorUnit.get_all_by_detector(
                self.client,
                self.detector_id,
                self._get_param("deviceMappingSnapshotAt"),
            )
        self._check_resp(resp, warning="Failed to retrieve module mapping")
        for irrelevant_key in ("detector", "detector_type", "flg_available"):
            for pdu in resp["data"]:
                del pdu[irrelevant_key]
        return resp["data"]

    @functools.cached_property
    def _karabo_da_to_float_uuid(self):
        return {pdu["karabo_da"]: pdu["float_uuid"] for pdu in self.pdus}

    @functools.cached_property
    def _karabo_da_to_id(self):
        return {pdu["karabo_da"]: pdu["id"] for pdu in self.pdus}

    def flush_pdu_mapping(self):
        for attr in ("pdus", "_karabo_da_to_float_uuid", "_karabo_da_to_id"):
            if hasattr(self, attr):
                delattr(self, attr)
        self._set_param("moduleId", "")

    @utils.threadsafe_cache
    def calibration_id(self, calibration_name: str):
        with self.api_lock:
            resp = Calibration.get_by_name(self.client, calibration_name)
        self._check_resp(
            resp,
            CalibrationNotFound,
            f"Calibration type {calibration_name} not found!"
        )
        return resp["data"]["id"]

    def condition_ids(self, pdu, condition):
        # note: do not cache, let CalCat search new condition IDs as they may differ
        condition_with_detector = copy.copy(condition)
        condition_with_detector["Detector UUID"] = pdu
        self.device.log.DEBUG(f"Look for condition: {condition_with_detector}")
        with self.api_lock:
            resp = self.client.search_possible_conditions_from_dict(
                "", condition_with_detector.encode()
            )
        self._check_resp(
            resp,
            ConditionNotFound,
            f"Failed to find condition {condition} for pdu {pdu}",
        )
        return [d["id"] for d in resp["data"]]

    def constant_ids(self, calibration_id, condition_ids):
        with self.api_lock:
            resp = CalibrationConstant.get_all_by_conditions(
                self.client,
                calibration_id=calibration_id,
                detector_type_id=self.detector_type_id,
                condition_ids=condition_ids,
            )
        self._check_resp(resp, warning="Failed to retrieve constant ID")
        return [d["id"] for d in resp["data"]]

    def get_constant_version(self, constant):
        karabo_da = self._get_param("karaboDa")
        self.device.log_status_info(f"Attempting to find {constant} for {karabo_da}")

        if karabo_da not in self._karabo_da_to_float_uuid:
            self.device.log_status_warn(
                f"Module {karabo_da} not found in mapping, check configuration!"
            )
            raise ModuleNotFound(f"Module map did not include {karabo_da}")
        self._set_param("moduleId", str(self._karabo_da_to_id[karabo_da]))

        if isinstance(constant, str):
            constant = self._constant_enum_class[constant]

        calibration_id = self.calibration_id(constant.name)
        self._set_status(constant, "calibrationId", calibration_id)

        condition = self._constants_need_conditions[constant]()
        condition_ids = self.condition_ids(
            self._karabo_da_to_float_uuid[karabo_da], condition
        )
        self._set_status(constant, "conditionIds", condition_ids)

        constant_ids = self.constant_ids(
            calibration_id=calibration_id, condition_ids=condition_ids
        )
        self._set_status(constant, "constantIds", constant_ids)

        with self.api_lock:
            resp = CalibrationConstantVersion.get_closest_by_time(
                self.client,
                calibration_constant_ids=constant_ids,
                physical_detector_unit_id=self._karabo_da_to_id[karabo_da],
                event_at=self._get_param("constantVersionEventAt"),
                snapshot_at=None,
            )
        self._check_resp(resp, warning="Failed to find calibration constant version")
        # note: could consider adding end date
        timestamp = resp["data"]["begin_validity_at"]
        self._set_status(constant, "beginValidityAt", timestamp)
        self._set_status(constant, "constantVersionId", resp["data"]["id"])
        self._set_status(
            constant,
            "usedConditionId",
            resp["data"]["calibration_constant"]["condition_id"],
        )
        self._set_status(
            constant, "usedConstantId", resp["data"]["calibration_constant"]["id"]
        )

        file_path = (
            self.caldb_store / resp["data"]["path_to_file"] / resp["data"]["file_name"]
        )
        self._set_status(constant, "dataFilePath", str(file_path))
        self._set_status(constant, "dataSetName", resp["data"]["data_set_name"])

        with timing.Stopwatch() as stopwatch:
            constant_data = _read_dataset_externally(
                file_path, resp["data"]["data_set_name"]
            )
        self._set_status(constant, "constantLoadingTime", stopwatch.elapsed)

        with self.cached_constants_lock:
            self.cached_constants[constant] = constant_data
        self.device.log_status_info(f"Done finding {constant.name} for {karabo_da}")

        return constant_data

    def get_constant_from_constant_version_id(self, constant):
        # TODO: warn if PDU or constant type does not match
        # TODO: warn if result is list (happens for empty version ID)
        constant_version_id = self.device.get(
            f"foundConstants.{constant.name}.constantVersionId"
        )

        with self.api_lock:
            resp = CalibrationConstantVersion.get_by_id(
                self.client, constant_version_id
            )
        self._check_resp(resp, warning="Failed to find calibration constant version")
        file_path = (
            self.caldb_store / resp["data"]["path_to_file"] / resp["data"]["file_name"]
        )
        self._set_status(constant, "dataFilePath", str(file_path))
        self._set_status(constant, "dataSetName", resp["data"]["data_set_name"])

        with timing.Stopwatch() as stopwatch:
            constant_data = _read_dataset_externally(
                file_path, resp["data"]["data_set_name"]
            )
        self._set_status(constant, "constantLoadingTime", stopwatch.elapsed)

        with self.cached_constants_lock:
            self.cached_constants[constant] = constant_data
        self._set_status(constant, "beginValidityAt", resp["data"]["begin_at"])
        self._set_status(constant, "calibrationId", "manual override")
        self._set_status(constant, "usedConditionId", "manual override")
        self._set_status(constant, "usedConstantId", "manual override")
        self._set_status(constant, "constantVersionId", constant_version_id)

        return constant_data

    def get_constant_from_file(self, constant):
        with timing.Stopwatch() as stopwatch:
            constant_data = _read_dataset_externally(
                self.device.get(f"foundConstants.{constant.name}.dataFilePath"),
                self.device.get(f"foundConstants.{constant.name}.dataSetName"),
            )
        self._set_status(constant, "constantLoadingTime", stopwatch.elapsed)

        with self.cached_constants_lock:
            self.cached_constants[constant] = constant_data
        self._set_status(constant, "beginValidityAt", "manual override")
        self._set_status(constant, "calibrationId", "manual override")
        self._set_status(constant, "usedConditionId", "manual override")
        self._set_status(constant, "usedConstantId", "manual override")
        self._set_status(constant, "constantVersionId", "manual override")

        return constant_data

    def flush_constants(self, constants=None, preserve_fields=None):
        if preserve_fields is None:
            preserve_fields = set()
        reset_fields = {
            "beginValidityAt",
            "calibrationId",
            "usedConditionId",
            "usedConstantId",
            "constantVersionId",
            "dataFilePath",
            "dataSetName",
        } - set(preserve_fields)
        if constants is None:
            constants = self._constant_enum_class
        with self.cached_constants_lock:
            for constant in constants:
                for field in reset_fields:
                    self._set_status(constant, field, "")
                if "state" not in preserve_fields:
                    self._set_status(constant, "state", "OFF")
                if constant in self.cached_constants:
                    del self.cached_constants[constant]

    def _check_resp(self, resp, exception=Exception, warning=None):
        # TODO: probably verify using "info" that exception is the right one
        to_raise = None
        if not resp["success"]:
            # TODO: probably more types of app_info errors?
            if resp["app_info"]:
                if "not found" in resp["info"]:
                    # this was likely the exception exception
                    to_raise = exception(resp["info"])
                else:
                    # but could also be authorization or similar issue
                    to_raise = CalibrationClientConfigError(resp["app_info"])
            to_raise = exception(resp["info"])
        if to_raise is not None:
            if warning is not None:
                self.device.log_status_warn(warning)
                raise to_raise


def _read_dataset_externally(file_path, data_set_name, append_data=True):
    def aux(queue):
        try:
            with h5py.File(file_path, "r") as fd:
                if append_data:
                    res = fd[data_set_name]["data"]
                else:
                    res = fd[data_set_name]
                queue.put(np.array(res))
        except Exception as ex:
            queue.put(ex)

    res_queue = multiprocessing.Queue()
    process = multiprocessing.Process(target=aux, args=(res_queue,))
    process.start()
    constant_data = res_queue.get()
    if isinstance(constant_data, Exception):
        raise constant_data
    return constant_data
