import enum

import numpy as np
from karabo.bound import (
    KARABO_CLASSINFO,
    OUTPUT_CHANNEL,
    OVERWRITE_ELEMENT,
    STRING_ELEMENT,
)

from .. import base_calcat, base_correction, base_kernel_runner, schemas, utils
from .._version import version as deviceVersion


class CorrectionFlags(enum.IntFlag):
    NONE = 0
    OFFSET = 1


class Constants(enum.Enum):
    Offset = enum.auto()


class DsscBaseRunner(base_kernel_runner.BaseKernelRunner):
    _corrected_axis_order = "fyx"
    _xp = None

    @property
    def map_shape(self):
        return (self.constant_memory_cells, self.pixels_y, self.pixels_x)

    @property
    def input_shape(self):
        return (self.frames, self.pixels_y, self.pixels_x)

    @property
    def processed_shape(self):
        return self.input_shape

    def __init__(
        self,
        pixels_x,
        pixels_y,
        frames,
        constant_memory_cells,
        output_data_dtype=np.float32,
    ):
        super().__init__(
            pixels_x,
            pixels_y,
            frames,
            constant_memory_cells,
            output_data_dtype,
        )
        self.offset_map = self._xp.empty(self.map_shape, dtype=np.float32)
        self.processed_data = self._xp.empty(
            self.processed_shape, dtype=output_data_dtype
        )

    def flush_buffers(self, constants):
        if Constants.Offset in constants:
            self.offset_map.fill(0)

    @property
    def preview_data_views(self):
        return (
            self.input_data,
            self.processed_data,
        )

    def load_offset_map(self, offset_map):
        # can have an extra dimension for some reason
        if len(offset_map.shape) == 4:  # old format (see offsetcorrection_dssc.py)?
            offset_map = offset_map[..., 0]
        # shape (now): x, y, memory cell
        offset_map = self._xp.asarray(offset_map, dtype=np.float32).transpose()
        self.offset_map[:] = offset_map[:self.constant_memory_cells]


class DsscCpuRunner(DsscBaseRunner):
    _gpu_based = False
    _xp = np

    def _post_init(self):
        self.input_data = None
        self.cell_table = None
        from ..kernels import dssc_cython

        self.correction_kernel = dssc_cython.correct

    def load_data(self, image_data, cell_table):
        self.input_data = image_data.astype(np.uint16, copy=False)
        self.cell_table = cell_table.astype(np.uint16, copy=False)

    def correct(self, flags):
        self.correction_kernel(
            self.input_data,
            self.cell_table,
            flags,
            self.offset_map,
            self.processed_data,
        )


class DsscGpuRunner(DsscBaseRunner):
    _gpu_based = True

    def _pre_init(self):
        import cupy as cp

        self._xp = cp

    def _post_init(self):
        self.input_data = self._xp.empty(self.input_shape, dtype=np.uint16)
        self.cell_table = self._xp.empty(self.frames, dtype=np.uint16)
        self.block_shape = (1, 1, 64)
        self.grid_shape = utils.grid_to_cover_shape_with_blocks(
            self.input_shape, self.block_shape
        )
        self.correction_kernel = self._xp.RawModule(
            code=base_kernel_runner.get_kernel_template("dssc_gpu.cu").render(
                {
                    "pixels_x": self.pixels_x,
                    "pixels_y": self.pixels_y,
                    "frames": self.frames,
                    "constant_memory_cells": self.constant_memory_cells,
                    "output_data_dtype": utils.np_dtype_to_c_type(
                        self.output_data_dtype
                    ),
                    "corr_enum": utils.enum_to_c_template(CorrectionFlags),
                }
            )
        ).get_function("correct")

    def load_data(self, image_data, cell_table):
        self.input_data.set(image_data)
        self.cell_table.set(cell_table)

    def correct(self, flags):
        self.correction_kernel(
            self.grid_shape,
            self.block_shape,
            (
                self.input_data,
                self.cell_table,
                np.uint8(flags),
                self.offset_map,
                self.processed_data,
            ),
        )


class DsscCalcatFriend(base_calcat.BaseCalcatFriend):
    _constant_enum_class = Constants

    @property
    def _constants_need_conditions(self):
        return {Constants.Offset: self.dark_condition}

    @staticmethod
    def add_schema(schema):
        super(DsscCalcatFriend, DsscCalcatFriend).add_schema(schema, "DSSC-Type")
        (
            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.memoryCells")
            .setNewDefaultValue(400)
            .commit(),

            OVERWRITE_ELEMENT(schema)
            .key("constantParameters.biasVoltage")
            .setNewDefaultValue(100)  # TODO: proper
            .commit()
        )

        base_calcat.add_status_schema_from_enum(schema, Constants)

    def dark_condition(self):
        res = base_calcat.OperatingConditions()
        res["Memory cells"] = self._get_param("memoryCells")
        res["Sensor Bias Voltage"] = self._get_param("biasVoltage")
        res["Pixels X"] = self._get_param("pixelsX")
        res["Pixels Y"] = self._get_param("pixelsY")
        return res


@KARABO_CLASSINFO("DsscCorrection", deviceVersion)
class DsscCorrection(base_correction.BaseCorrection):
    # subclass *must* set these attributes
    _base_output_schema = schemas.xtdf_output_schema()
    _correction_flag_class = CorrectionFlags
    _correction_steps = (("offset", CorrectionFlags.OFFSET, {Constants.Offset}),)
    _calcat_friend_class = DsscCalcatFriend
    _constant_enum_class = Constants
    _preview_outputs = ["outputRaw", "outputCorrected"]

    @staticmethod
    def expectedParameters(expected):
        (
            OUTPUT_CHANNEL(expected)
            .key("dataOutput")
            .dataSchema(DsscCorrection._base_output_schema)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.frames")
            .setNewDefaultValue(400)
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("dataFormat.outputAxisOrder")
            .setNewDefaultValue("fyx")
            .commit(),

            OVERWRITE_ELEMENT(expected)
            .key("preview.selectionMode")
            .setNewDefaultValue("pulse")
            .commit(),
        )
        DsscCalcatFriend.add_schema(expected)
        base_correction.add_addon_nodes(expected, DsscCorrection)
        base_correction.add_correction_step_schema(
            expected,
            DsscCorrection._correction_steps,
        )
        base_correction.add_preview_outputs(expected, DsscCorrection._preview_outputs)
        (
            # support both CPU and GPU kernels
            STRING_ELEMENT(expected)
            .key("kernelType")
            .tags("managed")
            .assignmentOptional()
            .defaultValue(base_kernel_runner.KernelRunnerTypes.GPU.name)
            .options(
                ",".join(
                    kernel_type.name
                    for kernel_type in base_kernel_runner.KernelRunnerTypes
                )
            )
            .reconfigurable()
            .commit(),
        )

    @property
    def input_data_shape(self):
        return (
            self.get("dataFormat.frames"),
            1,
            self.get("dataFormat.pixelsY"),
            self.get("dataFormat.pixelsX"),
        )

    @property
    def _kernel_runner_class(self):
        kernel_type = base_kernel_runner.KernelRunnerTypes[
            self.unsafe_get("kernelType")
        ]
        if kernel_type is base_kernel_runner.KernelRunnerTypes.CPU:
            return DsscCpuRunner
        else:
            return DsscGpuRunner

    def process_data(
        self,
        data_hash,
        metadata,
        source,
        train_id,
        image_data,
        cell_table,
        pulse_table,
    ):
        if self._frame_filter is not None:
            try:
                cell_table = cell_table[self._frame_filter]
                pulse_table = pulse_table[self._frame_filter]
                image_data = image_data[self._frame_filter]
            except IndexError:
                self.log_status_warn(
                    "Failed to apply frame filter, please check that it is valid!"
                )
                return

        try:
            # drop the singleton dimension for kernel runner
            self.kernel_runner.load_data(image_data[:, 0], cell_table.ravel())
        except ValueError as e:
            self.log_status_warn(f"Failed to load data: {e}")
            return
        except Exception as e:
            self.log_status_warn(f"Unknown exception when loading data to GPU: {e}")

        buffer_handle, buffer_array = self._shmem_buffer.next_slot()
        self.kernel_runner.correct(self._correction_flag_enabled)
        for addon in self._enabled_addons:
            addon.post_correction(
                self.kernel_runner.processed_data, cell_table, pulse_table, data_hash
            )
        self.kernel_runner.reshape(
            output_order=self.unsafe_get("dataFormat.outputAxisOrder"),
            out=buffer_array,
        )
        with self.warning_context(
            "processingState", base_correction.WarningLampType.PREVIEW_SETTINGS
        ) as warn:
            if self._correction_flag_enabled != self._correction_flag_preview:
                self.kernel_runner.correct(self._correction_flag_preview)
            (
                preview_slice_index,
                preview_cell,
                preview_pulse,
            ), preview_warning = utils.pick_frame_index(
                self.unsafe_get("preview.selectionMode"),
                self.unsafe_get("preview.index"),
                cell_table,
                pulse_table,
            )
            if preview_warning is not None:
                warn(preview_warning)
            preview_raw, preview_corrected = self.kernel_runner.compute_previews(
                preview_slice_index,
            )

        if self._use_shmem_handles:
            data_hash.set(self._image_data_path, buffer_handle)
            data_hash.set("calngShmemPaths", [self._image_data_path])
        else:
            data_hash.set(self._image_data_path, buffer_array)
            data_hash.set("calngShmemPaths", [])

        data_hash.set(self._cell_table_path, cell_table[:, np.newaxis])
        data_hash.set(self._pulse_table_path, pulse_table[:, np.newaxis])

        self._write_output(data_hash, metadata)
        self._preview_friend.write_outputs(metadata, preview_raw, preview_corrected)

    def _load_constant_to_runner(self, constant, constant_data):
        self.kernel_runner.load_offset_map(constant_data)
