from karabo.bound import (
    BOOL_ELEMENT,
    FLOAT_ELEMENT,
    NODE_ELEMENT,
    OUTPUT_CHANNEL,
    STRING_ELEMENT,
    UINT32_ELEMENT,
    ChannelMetaData,
    Dims,
    Encoding,
    Hash,
    ImageData,
    Timestamp,
)

import numpy as np

from . import schemas, utils


class PreviewFriend:
    @staticmethod
    def add_schema(
        schema, node_path="preview", output_channels=None, create_node=False
    ):
        if output_channels is None:
            output_channels = ["output"]

        if create_node:
            (
                NODE_ELEMENT(schema)
                .key(node_path)
                .displayedName("Preview")
                .description(
                    "Output specifically intended for preview in Karabo GUI. Includes "
                    "some options for throttling and adjustments of the output data."
                )
                .commit(),
            )
        (
            BOOL_ELEMENT(schema)
            .key(f"{node_path}.flipSS")
            .displayedName("Flip SS")
            .description("Flip image data along slow scan axis.")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(schema)
            .key(f"{node_path}.flipFS")
            .displayedName("Flip FS")
            .description("Flip image data along fast scan axis.")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            UINT32_ELEMENT(schema)
            .key(f"{node_path}.downsamplingFactor")
            .displayedName("Factor")
            .description(
                "If greater than 1, the preview image will be downsampled by this "
                "factor before sending. This is mostly to save bandwidth in case GUI "
                "updates start lagging."
            )
            .assignmentOptional()
            .defaultValue(1)
            .options("1,2,4,8")
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{node_path}.downsamplingFunction")
            .displayedName("Function")
            .description("Reduction function used during downsampling.")
            .assignmentOptional()
            .defaultValue("nanmax")
            .options("nanmax,nanmean,nanmin,nanmedian")
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(schema)
            .key(f"{node_path}.replaceNanWith")
            .displayedName("NaN replacement")
            .description(
                "Displaying images in KaraboGUI seems to not go well when there are "
                "NaN values in data. And there will be with bad pixel masking or just "
                "geometry space between modules. NaN values get replaced with this "
                "value to get around this; choose a value which clearly stands out "
                "from the image data you want to see."
            )
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .commit(),
        )
        for channel in output_channels:
            (
                OUTPUT_CHANNEL(schema)
                .key(f"{node_path}.{channel}")
                .dataSchema(schemas.preview_schema(wrap_image_in_imagedata=True))
                .description("See description of parent node, 'preview'.")
                .commit(),
            )

    def __init__(self, device, node_name="preview", output_channels=None):
        if output_channels is None:
            output_channels = ["output"]
        self.output_channels = output_channels
        self.device = device
        self.dev_id = self.device.getInstanceId()
        self.node_name = node_name
        self.outputs = [
            self.device.signalSlotable.getOutputChannel(f"{self.node_name}.{channel}")
            for channel in self.output_channels
        ]
        self.reconfigure(device._parameters)

    def write_outputs(self, timestamp, *datas, inplace=True):
        """Applies GUI-friendly preview settings (replace NaN, downsample, wrap as
        ImageData) and writes to output channels. Make sure datas length matches number
        of channels!"""
        if isinstance(timestamp, Hash):
            timestamp = Timestamp.fromHashAttributes(
                timestamp.getAttributes("timestamp")
            )
        for data, output, channel_name in zip(
            datas, self.outputs, self.output_channels
        ):
            if self.downsampling_factor > 1:
                data = utils.downsample_2d(
                    data,
                    self.downsampling_factor,
                    reduction_fun=self.downsampling_function,
                )
            elif not inplace:
                data = data.copy()
            if self.flip_ss:
                data = np.flip(data, 0)
            if self.flip_fs:
                data = np.flip(data, 1)
            if isinstance(data, np.ma.MaskedArray):
                data, mask = data.data, data.mask | ~np.isfinite(data)
            else:
                mask = ~np.isfinite(data)
            data[mask] = self.nan_replacement
            mask = mask.astype(np.uint8)
            output_hash = Hash(
                "image.data",
                ImageData(
                    data,
                    Dims(*data.shape),
                    Encoding.GRAY,
                    bitsPerPixel=32,
                ),
                "image.mask",
                ImageData(
                    mask,
                    Dims(*mask.shape),
                    Encoding.GRAY,
                    bitsPerPixel=8,
                ),
            )
            output.write(
                output_hash,
                ChannelMetaData(
                    f"{self.dev_id}:{self.node_name}.{channel_name}",
                    timestamp,
                ),
                copyAllData=False,
            )
            output.update(safeNDArray=True)

    def reconfigure(self, conf):
        if conf.has(f"{self.node_name}.downsamplingFunction"):
            self.downsampling_function = getattr(
                np, conf[f"{self.node_name}.downsamplingFunction"]
            )
        if conf.has(f"{self.node_name}.downsamplingFactor"):
            self.downsampling_factor = conf[f"{self.node_name}.downsamplingFactor"]
        if conf.has(f"{self.node_name}.flipSS"):
            self.flip_ss = conf[f"{self.node_name}.flipSS"]
        if conf.has(f"{self.node_name}.flipFS"):
            self.flip_fs = conf[f"{self.node_name}.flipFS"]
        if conf.has(f"{self.node_name}.replaceNanWith"):
            self.nan_replacement = conf[f"{self.node_name}.replaceNanWith"]
