import natsort
from calngUtils.scene_utils import (
    BASE_INC,
    NARROW_INC,
    DisplayRoundedFloat,
    DisplayRow,
    EditableRow,
    HorizontalLayout,
    Space,
    VerticalLayout,
    Vline,
    boxed,
    recursive_editable,
    scene_generator,
    schema_to_hash,
    titled,
)
from karabo.common.scenemodel.api import (
    DetectorGraphModel,
    DeviceSceneLinkModel,
    DisplayCommandModel,
    DisplayLabelModel,
    DisplayStateColorModel,
    DisplayTextLogModel,
    LabelModel,
    LampModel,
    LineEditModel,
    NDArrayGraphModel,
    SceneTargetWindow,
    TableElementModel,
    TrendGraphModel,
    UnknownWidgetDataModel,
    VectorXYGraphModel,
    WebLinkModel,
)


@titled("Found constants", width=6 * NARROW_INC)
@boxed
class FoundConstantsColumn(VerticalLayout):
    def __init__(
        self,
        device_id,
        schema_hash,
        prefix="foundConstants",
    ):
        super().__init__(padding=0)
        for constant_name in schema_hash.get(prefix).getKeys():
            constant_row = HorizontalLayout(
                LabelModel(
                    text=constant_name,
                    width=6 * NARROW_INC,
                    height=NARROW_INC,
                ),
                LampModel(
                    width=NARROW_INC,
                    height=NARROW_INC,
                    keys=[f"{device_id}.{prefix}.{constant_name}.state"],
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.{prefix}.{constant_name}.beginValidityAt"],
                    width=8 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.{prefix}.{constant_name}.constantVersionId"],
                    width=3 * BASE_INC,
                    height=BASE_INC,
                ),
                padding=0,
            )
            self.children.append(constant_row)
        self.children.append(
            DeviceSceneLinkModel(
                text="Overrides",
                keys=[f"{device_id}.availableScenes"],
                target="constant_overrides",
                target_window=SceneTargetWindow.Dialog,
                width=8 * BASE_INC,
                height=BASE_INC,
            )
        )


class ConstantLoadedAmpeln(HorizontalLayout):
    def __init__(self, device_id, schema_hash, prefix="foundConstants"):
        super().__init__(padding=0)
        self.children.extend(
            [
                LampModel(
                    keys=[f"{device_id}.{prefix}.{key}.state"],
                    width=BASE_INC,
                    height=BASE_INC,
                )
                for key in schema_hash.get(prefix).getKeys()
            ]
        )


@titled("Manager status", width=6 * NARROW_INC)
@boxed
class ManagerDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        restart_button = DisplayCommandModel(
            keys=[f"{device_id}.restartServers"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        instantiate_button = DisplayCommandModel(
            keys=[f"{device_id}.startInstantiate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        managed_properties_link = DeviceSceneLinkModel(
            text="All managed properties",
            keys=[f"{device_id}.availableScenes"],
            target="browse_schema",
            target_window=SceneTargetWindow.Dialog,
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        docs_link = WebLinkModel(
            text="Documentation",
            width=7 * BASE_INC,
            height=BASE_INC,
            target="https://rtd.xfel.eu/docs/calng/en/latest/devices/#calibration-manager",
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    docs_link,
                    padding=0,
                ),
                HorizontalLayout(
                    restart_button,
                    instantiate_button,
                    padding=0,
                ),
                HorizontalLayout(
                    managed_properties_link,
                    padding=0,
                ),
                status_log,
            ]
        )


@titled("Device status", width=6 * NARROW_INC)
@boxed
class ProcessingDeviceStatus(VerticalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        rate = DisplayRoundedFloat(
            keys=[f"{device_id}.performance.rate"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        processing_time = DisplayRoundedFloat(
            keys=[f"{device_id}.performance.processingTime"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        tid = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        status_log = DisplayTextLogModel(
            keys=[f"{device_id}.status"],
            width=14 * BASE_INC,
            height=14 * BASE_INC,
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    tid,
                    padding=0,
                ),
                HorizontalLayout(
                    rate,
                    processing_time,
                    padding=0,
                ),
            ]
        )
        if schema_hash.has("warningLamps"):
            self.children.append(
                VerticalLayout(
                    children=[
                        HorizontalLayout(
                            LampModel(
                                keys=[f"{device_id}.{warning_lamp}"],
                                width=BASE_INC,
                                height=BASE_INC,
                            ),
                            LabelModel(
                                text=warning_lamp,
                                width=8 * BASE_INC,
                                height=BASE_INC,
                            ),
                        )
                        for warning_lamp in schema_hash.getAttribute(
                            "warningLamps", "defaultValue"
                        )
                    ],
                    padding=0,
                )
            )
        self.children.append(status_log)


class CompactCorrectionDeviceOverview(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__(padding=0)
        self.children.extend(
            [
                DeviceSceneLinkModel(
                    text=device_id.split("/")[-1],
                    keys=[f"{device_id}.availableScenes"],
                    target="overview",
                    target_window=SceneTargetWindow.Dialog,
                    width=5 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayStateColorModel(
                    show_string=True,
                    keys=[f"{device_id}.state"],
                    width=5 * BASE_INC,
                    height=BASE_INC,
                ),
            ]
        )
        for warning_lamp in schema_hash.getAttribute("warningLamps", "defaultValue"):
            self.children.append(
                LampModel(
                    keys=[f"{device_id}.{warning_lamp}"],
                    width=BASE_INC,
                    height=BASE_INC,
                )
            )
        self.children.extend(
            [
                DisplayRoundedFloat(
                    keys=[f"{device_id}.performance.rate"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.trainId"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
                ConstantLoadedAmpeln(device_id, schema_hash),
            ]
        )


class CompactDaqOverview(HorizontalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        self.children.extend(
            [
                DisplayLabelModel(
                    keys=[f"{device_id}.deviceId"],
                    width=8 * BASE_INC,
                    height=BASE_INC,
                    font_size=9,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.DataDispatcher.trainStride"],
                    width=4 * BASE_INC,
                    height=BASE_INC,
                ),
            ]
        )


@boxed
class CompactDeviceLinkList(VerticalLayout):
    def __init__(self, device_ids):
        super().__init__()
        self.children.extend(
            [
                HorizontalLayout(
                    DeviceSceneLinkModel(
                        text=device_id.split("/")[-1],
                        keys=[f"{device_id}.availableScenes"],
                        target="",
                        target_window=SceneTargetWindow.Dialog,
                        width=7 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DisplayStateColorModel(
                        show_string=True,
                        keys=[f"{device_id}.state"],
                        width=5 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                )
                for device_id in device_ids
            ]
        )


@titled("Assembler status", width=8 * NARROW_INC)
@boxed
class AssemblerDeviceStatus(VerticalLayout):
    def __init__(self, device_id):
        super().__init__(padding=0)
        name = DisplayLabelModel(
            keys=[f"{device_id}.deviceId"],
            width=14 * BASE_INC,
            height=BASE_INC,
        )
        state = DisplayStateColorModel(
            show_string=True,
            keys=[f"{device_id}.state"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        train_id = DisplayLabelModel(
            keys=[f"{device_id}.trainId"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        rate = DisplayRoundedFloat(
            keys=[f"{device_id}.outProcessing"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        tof = DisplayRoundedFloat(
            keys=[f"{device_id}.timeOfFlight"],
            width=7 * BASE_INC,
            height=BASE_INC,
        )
        self.children.extend(
            [
                name,
                HorizontalLayout(
                    state,
                    train_id,
                    padding=0,
                ),
                HorizontalLayout(
                    rate,
                    tof,
                    padding=0,
                ),
                DeviceSceneLinkModel(
                    text="I'm actually a TrainMatcher",
                    keys=[f"{device_id}.availableScenes"],
                    target="trainMatcherScene",
                    target_window=SceneTargetWindow.Dialog,
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
                LabelModel(
                    text="My geometry device:",
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),
                DisplayLabelModel(
                    keys=[f"{device_id}.geometryDevice"],
                    width=14 * BASE_INC,
                    height=BASE_INC,
                ),  # TODO: some day, get dynamic link to this friend
            ]
        )


@titled("Stats")
@boxed
class StatsBox(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__()
        self.children.extend(
            [
                DisplayRow(device_id, schema_hash, "rate", 5, 3),
                DisplayRow(device_id, schema_hash, "numPixelsIncluded", 5, 3),
            ]
        )


@titled("ROI selection")
@boxed
class RoiSelection(VerticalLayout):
    def __init__(self, image_path, roi_paths):
        super().__init__()
        self.children.append(
            UnknownWidgetDataModel(
                keys=([image_path] + list(roi_paths)),
                klass="RectRoiGraph",
                height=25 * BASE_INC,
                width=30 * BASE_INC,
            )
        )


@titled("ROI preview")
@boxed
class RoiBox(VerticalLayout):
    def __init__(self, device_id):
        super().__init__()
        self.children.append(
            DetectorGraphModel(
                keys=[f"{device_id}.output.schema.zoomImage"],
                height=17 * BASE_INC,
                width=20 * BASE_INC,
            ),
        )


@titled("Preview settings")
@boxed
class PreviewSettings(HorizontalLayout):
    def __init__(self, device_id, schema_hash, node_name="preview", extras=None):
        super().__init__()
        self.children.extend(
            [
                VerticalLayout(
                    EditableRow(
                        device_id,
                        schema_hash,
                        f"{node_name}.replaceNanWith",
                        6,
                        4,
                    ),
                    HorizontalLayout(
                        EditableRow(
                            device_id,
                            schema_hash,
                            f"{node_name}.flipSS",
                            3,
                            2,
                        ),
                        EditableRow(
                            device_id,
                            schema_hash,
                            f"{node_name}.flipFS",
                            3,
                            2,
                        ),
                        padding=0,
                    ),
                ),
                Vline(height=3 * BASE_INC),
                VerticalLayout(
                    LabelModel(
                        text="Image downsampling",
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    EditableRow(
                        device_id,
                        schema_hash,
                        f"{node_name}.downsamplingFactor",
                        5,
                        5,
                    ),
                    EditableRow(
                        device_id,
                        schema_hash,
                        f"{node_name}.downsamplingFunction",
                        5,
                        5,
                    ),
                ),
            ]
        )
        if extras is not None:
            self.children.append(Vline(height=3 * BASE_INC))
            self.children.extend(extras)


@titled("Histogram settings")
@boxed
class HistogramSettings(HorizontalLayout):
    def __init__(self, device_id, schema_hash):
        super().__init__()
        self.children.extend(
            [
                VerticalLayout(
                    LabelModel(text="Bins", width=3 * BASE_INC, height=BASE_INC),
                    DisplayCommandModel(
                        keys=[f"{device_id}.histogram.resetBins"],
                        width=4 * BASE_INC,
                        height=BASE_INC,
                    ),
                    EditableRow(
                        device_id, schema_hash, "histogram.resetBinsOnRoiChange", 5, 1
                    ),
                    EditableRow(device_id, schema_hash, "histogram.numBins", 3, 3),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(text="Range", width=4 * BASE_INC, height=BASE_INC),
                    EditableRow(device_id, schema_hash, "histogram.rangeMin", 4, 3),
                    EditableRow(device_id, schema_hash, "histogram.rangeMax", 4, 3),
                    EditableRow(
                        device_id,
                        schema_hash,
                        "histogram.automaticallyExpandRange",
                        4,
                        3,
                    ),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(text="Averaging", width=4 * BASE_INC, height=BASE_INC),
                    EditableRow(
                        device_id, schema_hash, "histogram.rollingWindowSize", 4, 3
                    ),
                    padding=0,
                ),
            ]
        )


class PreviewDisplayArea(VerticalLayout):
    def __init__(
        self,
        device_id,
        schema_hash,
        channel_name,
        data_width=None,
        mask_width=None,
        data_height=None,
        mask_height=None,
    ):
        super().__init__()
        try:
            class_id = schema_hash.getAttribute(
                f"{channel_name}.schema.image.data", "classId"
            )
        except RuntimeError:
            warning = f"{channel_name} in schema of {device_id} missing classId"
            print(warning)
            self.children.append(
                LabelModel(text=warning, width=10 * BASE_INC, height=BASE_INC)
            )
            return
        if class_id == "ImageData":
            self.children.extend(
                [
                    DetectorGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.data"],
                        colormap="viridis",
                        width=60 * BASE_INC if data_width is None else data_width,
                        height=18 * BASE_INC if data_height is None else data_height,
                    ),
                    DetectorGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.mask"],
                        colormap="viridis",
                        width=60 * BASE_INC if mask_width is None else mask_width,
                        height=18 * BASE_INC if mask_height is None else mask_height,
                    ),
                ]
            )
        else:
            self.children.extend(
                [
                    NDArrayGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.data"],
                        width=60 * BASE_INC if data_width is None else data_width,
                        height=18 * BASE_INC if data_height is None else data_height,
                    ),
                    NDArrayGraphModel(
                        keys=[f"{device_id}.{channel_name}.schema.image.mask"],
                        width=60 * BASE_INC if mask_width is None else mask_width,
                        height=12 * BASE_INC if mask_height is None else mask_height,
                    ),
                ]
            )


@scene_generator
def correction_device_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    main_overview = HorizontalLayout(
        ProcessingDeviceStatus(device_id, schema_hash),
        VerticalLayout(
            recursive_editable(
                device_id,
                schema_hash,
                "constantParameters",
            ),
            DisplayCommandModel(
                keys=[f"{device_id}.loadMostRecentConstants"],
                width=10 * BASE_INC,
                height=BASE_INC,
            ),
        ),
        FoundConstantsColumn(device_id, schema_hash),
        recursive_editable(
            device_id,
            schema_hash,
            "corrections",
            max_depth=2,
        ),
    )
    return VerticalLayout(
        main_overview,
        LabelModel(
            text="Preview (corrected):",
            width=20 * BASE_INC,
            height=BASE_INC,
        ),
        PreviewDisplayArea(device_id, schema_hash, "preview.outputCorrected"),
        *(
            DeviceSceneLinkModel(
                text=f"Preview: {channel}",
                keys=[f"{device_id}.availableScenes"],
                target=f"preview:preview.{channel}",
                target_window=SceneTargetWindow.Dialog,
                width=16 * BASE_INC,
                height=BASE_INC,
            )
            for channel in schema_hash.get("preview").getKeys()
            if schema_hash.hasAttribute(f"preview.{channel}", "classId")
            and schema_hash.getAttribute(f"preview.{channel}", "classId")
            == "OutputChannel"
        ),
    )


@scene_generator
def lpdmini_splitter_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        ProcessingDeviceStatus(device_id, schema_hash),
        DisplayRow(device_id, schema_hash, "outputDataShape", 6, 8),
    )


@scene_generator
def correction_device_preview(device_id, schema, preview_channel):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        LabelModel(
            text=f"Preview: {preview_channel}",
            width=10 * BASE_INC,
            height=BASE_INC,
        ),
        PreviewSettings(
            device_id,
            schema_hash,
            extras=[
                VerticalLayout(
                    EditableRow(device_id, schema_hash, "preview.index", 8, 4),
                    EditableRow(device_id, schema_hash, "preview.selectionMode", 8, 4),
                )
            ],
        ),
        PreviewDisplayArea(device_id, schema_hash, preview_channel),
    )


@scene_generator
def correction_device_constant_overrides(device_id, schema, prefix="foundConstants"):
    schema_hash = schema_to_hash(schema)

    return HorizontalLayout(
        children=[
            VerticalLayout(
                VerticalLayout(
                    LabelModel(text="constant", width=6 * BASE_INC, height=BASE_INC),
                    LabelModel(text="timestamp", width=6 * BASE_INC, height=BASE_INC),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(
                        text="constant version ID",
                        width=6 * BASE_INC,
                        height=BASE_INC,
                    ),
                    Space(width=6 * BASE_INC, height=BASE_INC),
                    padding=0,
                ),
                VerticalLayout(
                    LabelModel(
                        text="data file name", width=6 * BASE_INC, height=BASE_INC
                    ),
                    LabelModel(
                        text="dataset name", width=6 * BASE_INC, height=BASE_INC
                    ),
                    Space(width=6 * BASE_INC, height=BASE_INC),
                    padding=0,
                ),
            )
        ]
        + [
            VerticalLayout(
                VerticalLayout(
                    HorizontalLayout(
                        LampModel(
                            keys=[f"{device_id}.{prefix}.{constant}.state"],
                            width=BASE_INC,
                            height=BASE_INC,
                        ),
                        LabelModel(
                            text=constant,
                            width=7 * BASE_INC,
                            height=BASE_INC,
                        ),
                    ),
                    DisplayLabelModel(
                        keys=[f"{device_id}.{prefix}.{constant}.beginValidityAt"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                        font_size=9,
                    ),
                    padding=0,
                ),
                DisplayCommandModel(
                    keys=[f"{device_id}.{prefix}.{constant}.loadMostRecent"],
                    width=8 * BASE_INC,
                    height=BASE_INC,
                ),
                VerticalLayout(
                    LineEditModel(
                        klass="EditableLineEdit",
                        keys=[f"{device_id}.{prefix}.{constant}.constantVersionId"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DisplayCommandModel(
                        keys=[
                            f"{device_id}.{prefix}.{constant}.overrideConstantFromVersion"
                        ],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                ),
                VerticalLayout(
                    LineEditModel(
                        klass="EditableLineEdit",
                        keys=[f"{device_id}.{prefix}.{constant}.dataFilePath"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    LineEditModel(
                        klass="EditableLineEdit",
                        keys=[f"{device_id}.{prefix}.{constant}.dataSetName"],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    DisplayCommandModel(
                        keys=[
                            f"{device_id}.{prefix}.{constant}.overrideConstantFromFile"
                        ],
                        width=8 * BASE_INC,
                        height=BASE_INC,
                    ),
                    padding=0,
                ),
            )
            for constant in schema_hash.get(prefix).getKeys()
        ]
    )


@scene_generator
def roitool_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            RoiSelection(f"{device_id}.output.schema.fullImage", [f"{device_id}.roi"]),
            VerticalLayout(
                RoiBox(device_id),
                StatsBox(device_id, schema_hash),
                HistogramSettings(device_id, schema_hash),
            ),
        ),
        VectorXYGraphModel(
            keys=[
                f"{device_id}.output.schema.histogram.x",
                f"{device_id}.output.schema.histogram.y",
                f"{device_id}.output.schema.histogram.yMean",
            ],
            height=16 * BASE_INC,
            width=52 * BASE_INC,
        ),
    )


@scene_generator
def manager_device_overview(
    manager_device_id,
    manager_device_schema,
    correction_device_ids,
    correction_device_schema,
    daq_device_ids,
    domain_device_ids,
):
    mds_hash = schema_to_hash(manager_device_schema)
    cds_hash = schema_to_hash(correction_device_schema)

    data_throttling_children = [
        LabelModel(
            text="Frame filter",
            width=11 * BASE_INC,
            height=BASE_INC,
        ),
        EditableRow(
            manager_device_id,
            mds_hash,
            "managedKeys.frameFilter.type",
            7,
            4,
        ),
        EditableRow(
            manager_device_id,
            mds_hash,
            "managedKeys.frameFilter.spec",
            7,
            4,
        ),
    ]

    if "managedKeys.daqTrainStride" in mds_hash:
        # Only add DAQ train stride if present on the schema, may be
        # disabled on the manager.
        data_throttling_children.insert(
            0,
            EditableRow(
                manager_device_id,
                mds_hash,
                "managedKeys.daqTrainStride",
                7,
                4,
            ),
        )

    return VerticalLayout(
        HorizontalLayout(
            ManagerDeviceStatus(manager_device_id),
            VerticalLayout(
                recursive_editable(
                    manager_device_id,
                    mds_hash,
                    "managedKeys.constantParameters",
                ),
                DisplayCommandModel(
                    keys=[f"{manager_device_id}.managedKeys.loadMostRecentConstants"],
                    width=10 * BASE_INC,
                    height=BASE_INC,
                ),
                recursive_editable(
                    manager_device_id,
                    mds_hash,
                    "managedKeys.preview",
                    max_depth=2,
                ),
                titled("Data throttling")(boxed(VerticalLayout))(
                    children=data_throttling_children,
                    padding=0,
                ),
            ),
            recursive_editable(
                manager_device_id,
                mds_hash,
                "managedKeys.corrections",
                max_depth=2,
            ),
        ),
        HorizontalLayout(
            titled("DAQs", width=8 * NARROW_INC)(boxed(VerticalLayout))(
                children=[Space(width=BASE_INC, height=BASE_INC)] * 2
                + [
                    CompactDaqOverview(device_id)
                    for device_id in natsort.natsorted(daq_device_ids)
                ],
                padding=0,
            ),
            titled("Correction devices", width=8 * NARROW_INC)(boxed(VerticalLayout))(
                children=[
                    HorizontalLayout(
                        VerticalLayout(
                            DeviceSceneLinkModel(
                                text="Performance dashboard",
                                keys=[f"{manager_device_id}.availableScenes"],
                                target="correction_performance_overview",
                                target_window=SceneTargetWindow.Dialog,
                                width=10 * BASE_INC,
                                height=BASE_INC,
                            ),
                            DeviceSceneLinkModel(
                                text="Correction constant overview",
                                keys=[f"{manager_device_id}.availableScenes"],
                                target="correction_constant_overview",
                                target_window=SceneTargetWindow.Dialog,
                                width=10 * BASE_INC,
                                height=BASE_INC,
                            ),
                            padding=0,
                        ),
                        WebLinkModel(
                            text="Documentation",
                            width=6 * BASE_INC,
                            height=BASE_INC,
                            target="https://rtd.xfel.eu/docs/calng/en/latest/devices/#correction-devices",
                        ),
                        padding=0,
                    )
                ]
                + [
                    CompactCorrectionDeviceOverview(device_id, cds_hash)
                    for device_id in natsort.natsorted(correction_device_ids)
                ],
                padding=0,
            ),
            titled("Other devices managed")(CompactDeviceLinkList)(
                natsort.natsorted(
                    set(domain_device_ids)
                    - set(correction_device_ids)
                    - {manager_device_id}
                )
            ),
        ),
    )


@scene_generator
def correction_device_performance_dashboard(correction_device_ids):
    correction_device_ids = natsort.natsorted(correction_device_ids)
    return HorizontalLayout(
        titled("Correction device links")(CompactDeviceLinkList)(correction_device_ids),
        VerticalLayout(
            TrendGraphModel(
                keys=[
                    f"{device_id}.performance.rate"
                    for device_id in correction_device_ids
                ],
                width=30 * BASE_INC,
                height=15 * BASE_INC,
                x_label="timestamp",
                y_label="rate",
                y_units="Hz",
            ),
            TrendGraphModel(
                keys=[
                    f"{device_id}.performance.processingTime"
                    for device_id in correction_device_ids
                ],
                width=30 * BASE_INC,
                height=15 * BASE_INC,
                x_label="timestamp",
                y_label="processing per train",
                y_units="ms",
            ),
            TrendGraphModel(
                keys=[
                    f"{device_id}.performance.ratioOfRecentTrainsReceived"
                    for device_id in correction_device_ids
                ],
                width=30 * BASE_INC,
                height=15 * BASE_INC,
                x_label="timestamp",
                y_label="ratio of trains received",
                y_units="%",
            ),
        ),
    )


@scene_generator
def correction_constant_dashboard(
    correction_device_ids, correction_device_schema, prefix="foundConstants"
):
    correction_device_schema = schema_to_hash(correction_device_schema)
    constant_names = list(correction_device_schema.get(prefix).getKeys())
    correction_device_ids = natsort.natsorted(correction_device_ids)
    return VerticalLayout(
        HorizontalLayout(
            Space(width=6 * BASE_INC, height=BASE_INC),
            *(
                LabelModel(text=constant, width=7 * BASE_INC, height=BASE_INC)
                for constant in constant_names
            ),
            padding=0,
        ),
        *(
            HorizontalLayout(
                DeviceSceneLinkModel(
                    text=device_id.split("/")[-1],
                    keys=[f"{device_id}.availableScenes"],
                    width=6 * BASE_INC,
                    height=BASE_INC,
                ),
                *(
                    DisplayLabelModel(
                        keys=[f"{device_id}.{prefix}.{constant}.beginValidityAt"],
                        width=7 * BASE_INC,
                        height=BASE_INC,
                        font_size=9,
                    )
                    for constant in constant_names
                ),
                padding=0,
            )
            for device_id in correction_device_ids
        ),
    )


@scene_generator
def detector_assembler_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            AssemblerDeviceStatus(device_id),
            PreviewSettings(device_id, schema_hash),
        ),
        PreviewDisplayArea(
            device_id,
            schema_hash,
            "preview.output",
            data_width=40 * BASE_INC,
            data_height=40 * BASE_INC,
            mask_width=40 * BASE_INC,
            mask_height=15 * BASE_INC,
        ),
    )


@scene_generator
def gotthard2_assembler_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            AssemblerDeviceStatus(device_id),
        ),
        PreviewDisplayArea(
            device_id,
            schema_hash,
            "output",
            data_width=40 * BASE_INC,
            data_height=40 * BASE_INC,
            mask_width=40 * BASE_INC,
            mask_height=15 * BASE_INC,
        ),
    )


@scene_generator
def condition_checker_overview(device_id, schema):
    schema_hash = schema_to_hash(schema)
    return VerticalLayout(
        HorizontalLayout(
            DisplayCommandModel(
                keys=[f"{device_id}.checkConditions"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
            DisplayCommandModel(
                keys=[f"{device_id}.updateConditions"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
            DisplayStateColorModel(
                keys=[f"{device_id}.conditionsMatch"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
        ),
        HorizontalLayout(
            DisplayCommandModel(
                keys=[f"{device_id}.startMonitoring"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
            DisplayCommandModel(
                keys=[f"{device_id}.stopMonitoring"],
                width=7 * BASE_INC,
                height=BASE_INC,
            ),
        ),
        EditableRow(device_id, schema_hash, "updateManagerOnMonitor", 13, 1),
        EditableRow(device_id, schema_hash, "loadConstantsOnMonitor", 13, 1),
        TableElementModel(
            keys=[f"{device_id}.keyMapping"],
            width=30 * BASE_INC,
            height=20 * BASE_INC,
        ),
    )
