import contextlib
import collections
import enum
import functools
import itertools

import numpy as np
from calngUtils import misc


class WarningContextSystem:
    """Helper object to trigger warning lamps based on different warning types in
    contexts. Intended use: something that is checked multiple times and is good or bad
    depending on last check (ex. per train).

    Warning model: a warning is generated within context (see __call_). Each warning
    context handles one schema key and one warning type for that key. One lamp
    aggregates multiple warning types; for each lamp, a set of currently active warnings
    decides the ultimate fate of the lamp.

    If a warning is issued within a context, the corresponding warning type is added to
    the set. If not, then the corresponding warning type is removed from the set."""

    def __init__(self, device, on_success=None, on_error=None):
        self.device = device
        # key -> set of warning types
        self.active_warnings = collections.defaultdict(lambda: set())
        # (key, warn type) -> True or False depending on last context
        self.triggered_warnings = collections.defaultdict(lambda: False)

        # key -> state string
        self.on_success = collections.defaultdict(lambda: "ON")
        if on_success is not None:
            self.on_success.update(on_success)
        self.on_error = collections.defaultdict(lambda: "ERROR")
        if on_error is not None:
            self.on_error.update(on_error)
        # TODO: support mapping different warning types to different Karabo states
        # (for colors per lamp depending on state signifier on active set)

    @contextlib.contextmanager
    def __call__(self, key, warn_type=None):
        """Will return a context handler object for tkey specified schema key and
        warning type. It is your responsibility to ensure that the key is valid. Using
        __call__ on the returned context handler will issue a warning and add the
        warning type to the set of active warnings for the key."""
        warn_fun = functools.partial(self.issue_warning, key, warn_type)
        try:
            yield warn_fun
        except Exception as ex:
            warn_fun(f"Unexpected exception: {ex}")
        finally:
            # warn_fun may have set trigger; we update active warnings out here
            active_set = self.active_warnings[key]
            if self.triggered_warnings[(key, warn_type)]:
                if warn_type not in active_set:
                    self.device.set(key, self.on_error[key])
                    active_set.add(warn_type)
                self.triggered_warnings[(key, warn_type)] = False
            else:
                # warning not triggered, so clear out active set
                if warn_type in active_set:
                    message = (
                        f"{key}: warning lifted!"
                        if warn_type is None
                        else f"{key} > {warn_type.name}: lifted!"
                    )
                    self.device.log.INFO(message)
                    self.device.set("status", message)
                    active_set.discard(warn_type)
                if not active_set:
                    self.device.set(key, self.on_success[key])

    def reset(self):
        self.active_warnings.clear()
        self.triggered_warnings.clear()

    def issue_warning(self, key, warn_type, text):
        self.triggered_warnings[(key, warn_type)] = True
        # already active, so don't print
        if warn_type in self.active_warnings[key]:
            return
        message = (
            f"{key}: {text}"
            if warn_type is None
            else f"{key} > {warn_type.name}: {text}"
        )
        # TODO: make log + "status" behavior customizable
        self.device.log.WARN(message)
        self.device.set("status", message)


class PreviewIndexSelectionMode(enum.Enum):
    FRAME = "frame"
    CELL = "cell"
    PULSE = "pulse"


def pick_frame_index(selection_mode, index, cell_table, pulse_table):
    """When selecting a single frame to preview, an obvious question is whether the
    number the operator provides is a frame index, a cell ID, or a pulse ID. This
    function allows any of the three, translating into frame index.

    Indices below zero are special values and thus returned directly.

    Returns: (frame index, cell ID, pulse ID), optional warning"""

    if index < 0:
        return (index, index, index), None

    warning = None
    selection_mode = PreviewIndexSelectionMode(selection_mode)

    if selection_mode is PreviewIndexSelectionMode.FRAME:
        if index < cell_table.size:
            frame_index = index
        else:
            warning = (
                f"Frame index {index} out of range for cell table of length "
                f"{len(cell_table)}. Will use frame index 0 instead."
            )
            frame_index = 0
    else:
        if selection_mode is PreviewIndexSelectionMode.CELL:
            index_table = cell_table
        else:
            index_table = pulse_table
        found = np.where(index_table == index)[0]
        if len(found) == 1:
            frame_index = found[0]
        elif len(found) == 0:
            frame_index = 0
            warning = (
                f"{selection_mode.value.capitalize()} ID {index} not found in "
                f"{selection_mode.name} table. Will use frame {frame_index}, which "
                f"corresponds to {selection_mode.name} {index_table[0]} instead."
            )
        elif len(found) > 1:
            warning = (
                f"{selection_mode.value.capitalize()} ID {index} was not unique in "
                f"{selection_mode} table. Will use first occurrence out of "
                f"{len(found)} occurrences (frame {frame_index})."
            )

    cell_id = cell_table[frame_index]
    pulse_id = pulse_table[frame_index]
    return (frame_index, cell_id, pulse_id), warning


_np_typechar_to_c_typestring = {
    "?": "bool",
    "B": "unsigned char",
    "D": "double complex",
    "F": "float complex",
    "G": "long double complex",
    "H": "unsigned short",
    "I": "unsigned int",
    "L": "unsigned long",
    "Q": "unsigned long long",
    "b": "char",
    "d": "double",
    "e": "half",  # warning: only in CUDA with special support
    "f": "float",
    "g": "long double",
    "h": "short",
    "i": "int",
    "l": "long",
    "q": "long long",
}


def np_dtype_to_c_type(dtype):
    as_char = np.sctype2char(dtype)
    return _np_typechar_to_c_typestring[as_char]


def enum_to_c_template(enum_class):
    res = [f"enum {enum_class.__name__} {{"]
    for field in enum_class:
        res.append(f"\t{field.name} = {field.value},")
    res.append("};")
    return "\n".join(res)


class StateContext:
    """What if device state was a stack?"""

    def __init__(self, device, new_state, revert_to=None):
        self.device = device
        if revert_to is None:
            revert_to = device.get("state")
        self.revert_to = revert_to
        self.new_state = new_state

    def __enter__(self):
        if self.new_state is not self.revert_to:
            self.device.updateState(self.new_state)

    def __exit__(self, t, v, tb):
        if self.new_state is not self.revert_to:
            self.device.updateState(self.revert_to)


class BadPixelValues(enum.IntFlag):
    """The European XFEL Bad Pixel Encoding

    Straight from pycalibration's enum.py"""

    OFFSET_OUT_OF_THRESHOLD = 2**0
    NOISE_OUT_OF_THRESHOLD = 2**1
    OFFSET_NOISE_EVAL_ERROR = 2**2
    NO_DARK_DATA = 2**3
    CI_GAIN_OUT_OF_THRESHOLD = 2**4
    CI_LINEAR_DEVIATION = 2**5
    CI_EVAL_ERROR = 2**6
    FF_GAIN_EVAL_ERROR = 2**7
    FF_GAIN_DEVIATION = 2**8
    FF_NO_ENTRIES = 2**9
    CI2_EVAL_ERROR = 2**10
    VALUE_IS_NAN = 2**11
    VALUE_OUT_OF_RANGE = 2**12
    GAIN_THRESHOLDING_ERROR = 2**13
    DATA_STD_IS_ZERO = 2**14
    ASIC_STD_BELOW_NOISE = 2**15
    INTERPOLATED = 2**16
    NOISY_ADC = 2**17
    OVERSCAN = 2**18
    NON_SENSITIVE = 2**19
    NON_LIN_RESPONSE_REGION = 2**20
    WRONG_GAIN_VALUE = 2**21
    NON_STANDARD_SIZE = 2**22


def downsample_2d(arr, factor, reduction_fun=np.nanmax):
    """Generalization of downsampling from FemDataAssembler

    Expects first two dimensions of arr to be multiple of 2 ** factor
    Useful if you're sitting at home and ssh connection is slow to get full-resolution
    previews."""

    for i in range(factor // 2):
        arr = reduction_fun(
            (
                arr[:-1:2],
                arr[1::2],
            ),
            axis=0,
        )
        arr = reduction_fun(
            (
                arr[:, :-1:2],
                arr[:, 1::2],
            ),
            axis=0,
        )
    return arr


def blocked_view_2d(A, blocks_ss, blocks_fs):
    assert A.shape[0] % blocks_ss == 0
    assert A.shape[1] % blocks_fs == 0
    return A.reshape(
        blocks_ss,
        A.shape[0] // blocks_ss,
        blocks_fs,
        A.shape[1] // blocks_fs,
    ).swapaxes(
        1,
        2,
    )


def quadrant_views(A):
    res = []
    for row in np.vsplit(A, 2):
        res.extend(np.hsplit(row, 2))
    return res


def cell_table_to_string(cell_table):
    return ",".join(map(str, cell_table.ravel())) + ","


def grid_to_cover_shape_with_blocks(full_shape, block_shape):
    return tuple(itertools.starmap(misc.ceil_div, zip(full_shape, block_shape)))


def apply_partial_lut(data, lut, mask, out, missing=np.nan):
    tmp = out.ravel()
    tmp[~mask] = data.ravel()[lut]
    tmp[mask] = missing
