import h5py
import numpy as np
import pathlib

from calng.corrections.AgipdCorrection import (
    AgipdCpuRunner,
    AgipdGpuRunner,
    Constants,
    CorrectionFlags,
)

output_dtype = np.float32
corr_dtype = np.float32
pixels_x = 512
pixels_y = 128
memory_cells = 352

raw_data = np.random.randint(
    low=0, high=2000, size=(memory_cells, 2, pixels_x, pixels_y), dtype=np.uint16
)
image_data = raw_data[:, 0]
raw_gain = raw_data[:, 1]
cell_table = np.arange(memory_cells, dtype=np.uint16)
np.random.shuffle(cell_table)

caldb_store = pathlib.Path("/gpfs/exfel/d/cal/caldb_store/xfel/cal")
caldb_prefix = caldb_store / "agipd-type/agipd_siv1_agipdv11_m305"

with h5py.File(caldb_prefix / "cal.1619543695.4679213.h5", "r") as fd:
    thresholds = np.array(fd["/AGIPD_SIV1_AGIPDV11_M305/ThresholdsDark/0/data"])
with h5py.File(caldb_prefix / "cal.1619543664.1545036.h5", "r") as fd:
    offset_map = np.array(fd["/AGIPD_SIV1_AGIPDV11_M305/Offset/0/data"])
with h5py.File(caldb_prefix / "cal.1615377705.8904035.h5", "r") as fd:
    slopes_pc_map = np.array(fd["/AGIPD_SIV1_AGIPDV11_M305/SlopesPC/0/data"])

kernel_runners = [
    runner_class(
        pixels_x,
        pixels_y,
        memory_cells,
        constant_memory_cells=memory_cells,
        output_data_dtype=output_dtype,
    )
    for runner_class in (AgipdCpuRunner, AgipdGpuRunner)
]


def thresholding_cpu(data, cell_table, thresholds):
    # get to memory_cell, x, y
    raw_gain = data[:, 1, ...].astype(corr_dtype, copy=False)
    # get to threshold, memory_cell, x, y
    thresholds = np.transpose(thresholds)[:, cell_table]
    res = np.zeros((memory_cells, pixels_x, pixels_y), dtype=np.uint8)
    res[raw_gain > thresholds[0]] = 1
    res[raw_gain > thresholds[1]] = 2
    return res


gain_map_cpu = thresholding_cpu(raw_data, cell_table, thresholds)


def corr_offset_cpu(data, cell_table, gain_map, offset):
    image_data = data[:, 0].astype(corr_dtype, copy=False)
    offset = np.transpose(offset)[:, cell_table]
    return (image_data - np.choose(gain_map, offset)).astype(output_dtype)


def corr_rel_gain_pc_cpu(data, cell_table, gain_map, slopes_pc):
    slopes_pc = slopes_pc.astype(np.float32, copy=False)
    pc_high_m = slopes_pc[0]
    pc_high_I = slopes_pc[1]
    pc_med_m = slopes_pc[3]
    pc_med_I = slopes_pc[4]
    frac_high_med = pc_high_m / pc_med_m
    md_additional_offset = pc_high_I - pc_med_I * frac_high_med
    rel_gain_map = np.ones((3, pixels_x, pixels_y, memory_cells), dtype=np.float32)
    rel_gain_map[0] = 1  # rel xray gain can come after
    rel_gain_map[1] = rel_gain_map[0] * np.transpose(frac_high_med)
    rel_gain_map[2] = rel_gain_map[1] * 4.48
    res = data[:, 0].astype(corr_dtype, copy=True)
    res *= np.choose(gain_map, np.transpose(rel_gain_map, (0, 3, 1, 2)))
    pixels_in_medium_gain = gain_map == 1
    res[pixels_in_medium_gain] += np.transpose(md_additional_offset, (0, 2, 1))[
        pixels_in_medium_gain
    ]
    return res


def test_thresholding():
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        runner.load_constant(Constants.ThresholdsDark, thresholds)
        runner.correct(np.uint8(CorrectionFlags.THRESHOLD))
    assert np.allclose(kernel_runners[0].gain_map, gain_map_cpu)
    assert np.allclose(kernel_runners[1].gain_map.get(), gain_map_cpu)


def test_offset():
    reference = corr_offset_cpu(raw_data, cell_table, gain_map_cpu, offset_map)
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        runner.load_constant(Constants.ThresholdsDark, thresholds)
        runner.load_constant(Constants.Offset, offset_map)
        # have to do thresholding, otherwise all is treated as high gain
        runner.correct(np.uint8(CorrectionFlags.THRESHOLD | CorrectionFlags.OFFSET))
        res = runner.reshape(runner._corrected_axis_order)
        assert np.allclose(res, reference), f"{runner.__class__}"


def test_rel_gain_pc():
    reference = corr_rel_gain_pc_cpu(raw_data, cell_table, gain_map_cpu, slopes_pc_map)
    for runner in kernel_runners:
        runner.load_data(raw_data, cell_table)
        runner.load_constant(Constants.ThresholdsDark, thresholds)
        runner.load_constant(Constants.SlopesPC, slopes_pc_map)
        runner.correct(
            np.uint8(CorrectionFlags.THRESHOLD | CorrectionFlags.REL_GAIN_PC)
        )
        res = runner.reshape(runner._corrected_axis_order)
        assert np.allclose(res, reference)
